
/*

Attempts to merge fragments of uuencoded files (read on `stdin') and
decode them.  The fragments must consist of whole lines and be in the
correct order.  They can be interleaved with arbitrary lines of text
which don't look too much like uuencoded data (see syntax below), and
which don't appear within about two lines of the end of a a uuencoded
file.

Lines recognised:

  B = begin <mode> <string>
  M = M<60chars> | M<61chars>
  E = end

State transitions

  Start : On
            B
              Output B line;
              --> Processing
            any
              skip

  Processing : On
  	         M
                   clear buffer ;
  		   Output M line
  	         E
  		   Output up to two lines from buffer;
  		   Output E line;
  		   --> Start
  	         any
  		   Add line to buffer
  		 
(I'm not sure what the last couple of lines before 'end' in a uuencoded
file are for.  The strategy I've adopted is that when `end' is
encountered the most recently read non-M lines are output, up to a
maximum of two of them)

*/


#include <stdio.h>
 
#define TRUE  1
#define FALSE 0


/* States */

#define START       0
#define PROCESSING  1
#define DONE        2

/* Tokens representing lines */

#define BEGIN   0
#define MIDDLE  1
#define END     2
#define ANY     3


/* Predicates on characters */

isOctal(ch)
char ch ;
{
  return '0' <= ch && ch <= '7' ;
}


char *file ;

/* Predicates on lines */

int isBegin(line)
char *line ;
{
  if(line [0] == 'b' && line [1] == 'e' && line [2] == 'g' &&
     line [3] == 'i' && line [4] == 'n' && line [5] == ' ') {
    if(isOctal(line [6]) && isOctal(line [7]) && isOctal(line [8]) &&
       line [9] == ' ') {
      file = line + 10 ;
      return 1 ;
    }
    if(isOctal(line [6]) && isOctal(line [7]) && isOctal(line [8]) &&
       isOctal(line [9]) && line [10] == ' ') {
      file = line + 11 ;
      return 1 ;
    }
  }
  return 0 ;
}

int isMiddle(line)
char *line ;
{
  int l ;

  return line [0] == 'M' && ((l = strlen(line)) == 61 || l == 62) ;
}

int isEnd(line)
char *line ;
{
  return line [0] == 'e' && line [1] == 'n' && line [2] == 'd' &&
         line [3] == (char) 0 ;
}


/* Get a line and classify it */

int getLine(line)
char *line ;
{
  if (gets(line)) {
    if (isMiddle(line))
      return MIDDLE ;
    else if (isBegin(line))
      return BEGIN ;
    else if (isEnd(line))
      return END ;
    else
      return ANY ;
  }
  else
    return EOF ;
}


/* Low level IO and process manipulation */

FILE *slave ;

putLine(line)
char *line ;
{
  fprintf(slave, "%s\n", line) ;
}

copyLine(lval, rval)
char *lval, *rval ;
{
  strcpy(lval, rval) ;
}

startProcessing(line)
char *line ;
{
  int chan [2] ;

  fprintf(stderr, "extracting %s\n", file) ;

  pipe(chan) ;
  if(fork(0) == 0) {
    /* close child process's reference to output end of pipe */
    close(chan [1]) ;

    /* connect input end of pipe to standard input */
    close(0) ;
    dup(chan [0]) ;
    close(chan [0]) ;

    /* run `uudecode' */
    execlp("/bin/uudecode", "uudecode",  (char *) 0) ;
  }
  /* close parent process's reference to input end of pipe */
  close(chan [0]) ;

  /* connect output end of pipe to `slave' file structure */
  slave = fdopen(chan [1], "w") ;

  putLine(line) ;
}

endProcessing()
{
  int status ;

  fclose(slave) ;
  wait(&status) ;
}

#define STACKMAX 2

main ()
{
  int state, stackSize, i ;
  char line [1024], stack [STACKMAX] [1024] ;

  state = START ;
  while(state != DONE) {
    switch (state) {
      case START :
	switch (getLine(line)) {
	  case BEGIN :
            startProcessing(line) ;
            stackSize = 0 ;
	    state = PROCESSING ;
	    break ;
          case EOF :
            state = DONE ;
            break ;
          default :
            ;
        }
        break ;
      case PROCESSING :
	switch (getLine(line)) {
          case MIDDLE :
            putLine(line) ;
            stackSize = 0 ;
            break ;
          case END :
            /* Output stack in bottom-up order */
            for (i = stackSize - 1; i >= 0; i--)
              putLine(stack [i]) ;
            putLine(line) ;
            endProcessing() ;
            state = START ;
            break ;
          case BEGIN :
            endProcessing() ;  /* missing `end' line error */
            startProcessing(line) ;
            stackSize = 0 ;
	    state = PROCESSING ;
	    break ;
          case EOF :
            endProcessing() ;  /* missing `end' line error */
            state = START ;
            break ;
          default :
            /* Push line onto stack (if I had time, I'd use circular buffer) */
            for (i = STACKMAX - 1; i > 0; i--)
              copyLine(stack [i], stack [i - 1]) ;
            copyLine(stack [0], line) ;
            if (stackSize < STACKMAX) stackSize++ ;
        }
    }
  }
}
      
